__author__ = 'Robbert Harms'
__date__ = "2014-06-19"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AbstractParameterProposal(object):
    """The parameter proposals are meant for use during sampling.

    Proposal functions are an essential part in sampling as they provide new candidate points for the sampler.

    The proposals allow for dynamic updates to better adapt to the landscape during sampling.
    """

    def is_symmetric(self):
        """Check if this proposal is symmetric. That is, if q(x|y) == q(y|x)."""
        return True

    def get_parameters(self):
        """The proposal parameters.

        Returns:
            list of ProposalParameter: the proposal parameter objects used by this proposal
        """

    def get_proposal_function(self):
        """Get the proposal function as a CL string. This should include include guards (#ifdef's).

        This should follow the signature:

        .. code-block: c

            mot_float_type <proposal_fname>(mot_float_type current, void* rng_data, <additional_parameters>)

        That is, it can have more than two parameter, but the first two are obligatory. The additional parameters
        are given by :meth:`get_parameters`.

        Returns:
            str: The cl function
        """

    def get_proposal_function_name(self):
        """Get the name of the proposal function call.

         This is used by the model builder to construct the call to the proposal function.

         Returns:
            str: name of the function
        """

    def get_proposal_logpdf_function(self):
        """Get the proposal pdf function as a CL string.

        This function is used if the proposal is not symmetric. The implementation should include include guards.

        This should follow the signature:

        .. code-block: c

            mot_float_type <proposal_pdf_fname>(mot_float_type proposal, mot_float_type current,
                                                <additional_parameters>)

        Returns:
            str: The proposal log pdf function as a CL string
        """

    def get_proposal_logpdf_function_name(self):
        """Get the name of the proposal logpdf function call.

         This is used by the model builder to construct the call to the proposal logpdf function.

         Returns:
            str: name of the function
        """

    def get_proposal_update_function(self):
        """Get the proposal update function to use for updating the adaptable parameters.

        Returns:
            ProposalUpdate: the proposal update function defining the update mechanism
        """


class ProposalUpdate(object):

    def get_update_function(self, proposal_parameters, address_space='private'):
        """Get the update function to update the proposal parameters of one of the proposals.

        Args:
            proposal_parameters (list of ProposalParameter): the list of proposal parameters to (possibly use) in the
                update function. It will only use the parameters that have ``adaptable`` set to True.
            address_space (str): the address space to use for the function parameters
        """

    def get_function_name(self, proposal_parameters):
        """Get the name of the proposal update function.

        Args:
            proposal_parameters (list of ProposalParameter): the list of proposal parameters to (possibly use) in the
                update function. It will only use the parameters that have ``adaptable`` set to True.

        Returns:
            str: the name of the function returned by :meth:`get_update_function`
        """

    def uses_parameter_variance(self):
        """Check if this proposal update function uses the parameter variance.

        If not, we will not provide it, this saves memory in the kernel.

        Returns:
            boolean: if this proposal update function uses the parameter variance
        """

    def uses_jump_counters(self):
        """Check if this proposal update function uses the jump counters (jump counter and acceptance counter).

        If not, we will not provide it, this saves memory in the kernel.

        Returns:
            boolean: if this proposal update function uses the jump counters
        """


class SimpleProposalUpdate(ProposalUpdate):

    def __init__(self, function_name, uses_parameter_variance=False, uses_jump_counters=True):
        """A simple proposal update function template.

        Args:
            function_name (str): the name of this proposal update function, try to choose an unique name.
            uses_parameter_variance (boolean): if this proposal requires the parameter variance. Enable if you need it
                 in your update function.
            uses_jump_counters (boolean): if this proposal uses jump counters for its workings. This is enabled
                by default. You can disable it for speed.
        """
        self._function_name = function_name
        self._uses_parameter_variance = uses_parameter_variance
        self._uses_jump_counters = uses_jump_counters

    def get_update_function(self, proposal_parameters, address_space='private'):
        return self._update_function_template('', proposal_parameters, address_space)

    def _update_function_template(self, function_body, proposal_parameters, address_space):
        params = ['{address_space} mot_float_type* const {name}'.format(address_space=address_space, name=p.name)
                  for p in proposal_parameters if p.adaptable]

        if self.uses_jump_counters():
            params.extend(['{address_space} uint* const sampling_counter'.format(address_space=address_space),
                           '{address_space} uint* const acceptance_counter'.format(address_space=address_space)])

        if self.uses_parameter_variance():
            params.append('mot_float_type parameter_variance'.format(address_space=address_space))

        return '''
            #ifndef {include_guard_name}
            #define {include_guard_name}

            mot_float_type {function_name}({params}){{
                {function_body}
            }}

            #endif //{include_guard_name}

        '''.format(include_guard_name='PROPOSAL_UPDATE_{}'.format(self._function_name.upper()),
                   function_name=self.get_function_name(proposal_parameters), params=', '.join(params),
                   function_body=function_body)

    def get_function_name(self, proposal_parameters):
        return 'proposal_update_{}_{}'.format(self._function_name,
                                              len([True for p in proposal_parameters if p.adaptable]))

    def uses_parameter_variance(self):
        return self._uses_parameter_variance

    def uses_jump_counters(self):
        return self._uses_jump_counters


class NoOptUpdateFunction(SimpleProposalUpdate):

    def __init__(self):
        """This is the no-operation update function. It does not update the proposal parameters."""
        super(NoOptUpdateFunction, self).__init__('no_opt')

    def get_update_function(self, proposal_parameters, address_space='private'):
        return self._update_function_template('', proposal_parameters, address_space)


class AcceptanceRateScaling(SimpleProposalUpdate):

    def __init__(self, target_acceptance_rate=0.44, batch_size=50, damping_factor=1):
        """Scales the proposal parameter (typically the std) such that it oscillates towards the chosen acceptance rate.

        This uses an scaling similar to the one in: "Examples of Adaptive MCMC",
        Gareth O. Roberts & Jeffrey S. Rosenthal (2009)

        This class implements the delta function as: :math:`\delta(n) = \sqrt{1 / (d*n)}`.
        Where n is the current batch index and d is the damping factor.

        With the default damping factor of 500, delta reaches a scaling of 0.01 in 20 batches. At a default batch size
        of 50 that amounts to 1000 samples.

        Args:
            target_acceptance_rate (float): the target acceptance rate between 0 and 1.
            batch_size (int): the size of the batches inbetween which we update the parameters
        """
        super(AcceptanceRateScaling, self).__init__('acceptance_rate_scaling')
        self._target_acceptance_rate = target_acceptance_rate
        self._batch_size = batch_size
        self._damping_factor = damping_factor

        if target_acceptance_rate > 1 or target_acceptance_rate < 0:
            raise ValueError('The target acceptance rate should be '
                             'within [0, 1], {} given.'.format(target_acceptance_rate))

    def get_update_function(self, proposal_parameters, address_space='private'):
        body = '''
            if(*sampling_counter % {batch_size} == 0){{

                mot_float_type delta = sqrt(1.0/({damping_factor} * (*sampling_counter / {batch_size})));

                if(*acceptance_counter / (mot_float_type){batch_size} > {target_ar}){{
                    *std *= exp(delta);
                }}
                else{{
                    *std /= exp(delta);
                }}

                *std = clamp(*std, (mot_float_type)1e-13, (mot_float_type)1e3);

                *acceptance_counter = 0;
            }}
        '''.format(batch_size=self._batch_size, target_ar=self._target_acceptance_rate,
                   damping_factor=self._damping_factor)
        return self._update_function_template(body, proposal_parameters, address_space)


class FSLAcceptanceRateScaling(SimpleProposalUpdate):

    def __init__(self, batch_size=50, min_val=1e-13, max_val=1e3):
        """An acceptance rate scaling algorithm found in a Neuroscience package called FSL.

        This scaling algorithm scales the std. by :math:`\sqrt(a/(n - a))` where a is the number of accepted samples
        in the last batch and n is the batch size. Its goal is to balance the acceptance rate at 0.5.

        So far, the author of this function in MOT has not been able to find theoretical support for this scaling
        algorithm. Please use this heuristic with caution.

        To prevent runaway proposal values we clamp the updated parameter value between a minimum and maximum value
        specified in the constructor.

        Args:
            batch_size (int): the size of the batches in between which we update the parameters
            min_val (float): the minimum value the parameter can take
            max_val (float): the maximum value the parameter can take
        """
        super(FSLAcceptanceRateScaling, self).__init__('fsl_acceptance_rate_scaling')
        self._batch_size = batch_size
        self._min_val = min_val
        self._max_val = max_val

    def get_update_function(self, proposal_parameters, address_space='private'):
        body = '''
            if(*sampling_counter == {batch_size}){{
                *std = clamp(*std * sqrt((mot_float_type)(*acceptance_counter + 1) /
                                         ({batch_size} - *acceptance_counter + 1)),
                             (mot_float_type){min_val},
                             (mot_float_type){max_val});

                *sampling_counter = 0;
                *acceptance_counter = 0;
            }}
        '''.format(batch_size=self._batch_size, min_val=self._min_val, max_val=self._max_val)
        return self._update_function_template(body, proposal_parameters, address_space)


class SingleComponentAdaptiveMetropolis(SimpleProposalUpdate):

    def __init__(self, waiting_period=100, scaling_factor=2.4, epsilon=1e-20):
        """Uses the Single Component Adaptive Metropolis (SCAM) scheme to update the proposals.

        This uses an scaling described in: "Componentwise adaptation for high dimensional MCMC",
        Heikki Haario, Eero Saksman and Johanna Tamminen (2005). That is, it updates the proposal standard deviation
        using the variance of the chain's history.

        Args:
            waiting_period (int): only start updating the proposal std. after this many draws.
            scaling_factor (float): the scaling factor to use (the parameter ``s`` in the paper referenced).
            epsilon (float): small number to prevent the std. from collapsing to zero.
        """
        super(SingleComponentAdaptiveMetropolis, self).__init__('scam', uses_parameter_variance=True)
        self._waiting_period = waiting_period
        self._scaling_factor = scaling_factor
        self._epsilon = epsilon

    def get_update_function(self, proposal_parameters, address_space='private'):
        body = '''
            if(*sampling_counter > {waiting_period}){{
                *std = {scaling_factor} * sqrt(parameter_variance) + {epsilon};
            }}
        '''.format(waiting_period=self._waiting_period, scaling_factor=self._scaling_factor, epsilon=self._epsilon)
        return self._update_function_template(body, proposal_parameters, address_space)


class ProposalParameter(object):

    def __init__(self, name, default_value, adaptable):
        """Container class for parameters of a proposal function.

        Args:
            default_value (double): the parameter value
            adaptable (boolean): if this parameter is adaptable during sampling

        Attributes:
            default_value (double): the parameter value
            adaptable (boolean): if this parameter is adaptable
        """
        self.name = name
        self.default_value = default_value
        self.adaptable = adaptable


class SimpleProposal(AbstractParameterProposal):

    def __init__(self, proposal_body, proposal_name, parameters, is_symmetric=True, logpdf_body='return 0;',
                 proposal_update_function=None):
        """Simple proposal template class.

        By default this assumes that the proposal you are generating is symmetric. If so, the proposal logpdf function
        can be reduced to a scalar since calculating it is unnecessary.

        Args:
            proposal_body (str): the body of the proposal code. Environment variables are ``current`` for the current
                position of this parameter and ``rng_data`` that can be used to generate random numbers.
            proposal_name (str): the name of this proposal
            parameters (list): the list of :class:`ProposalParameters` used by this proposal
            is_symmetric (boolean): if this proposal is symmetric, that is, if ``q(x|y) == q(y|x)``.
            logpdf_body (str): if the proposal is not symmetric we need a PDF function to calculate the probability of
                ``q(x|y)`` and ``q(y|x)``. It should return the log of the probability. If the proposal is symmetric
                this parameter need not be specified and defaults to returning a scalar.
            proposal_update_function (ProposalUpdate): the proposal update function to use. Defaults
                to :class:`SingleComponentAdaptiveMetropolis`.
        """
        self._proposal_body = proposal_body
        self._proposal_name = proposal_name
        self._parameters = parameters
        self._is_symmetric = is_symmetric
        self._logpdf_body = logpdf_body
        self._proposal_update_function = proposal_update_function or SingleComponentAdaptiveMetropolis()

    def is_symmetric(self):
        return self._is_symmetric

    def get_parameters(self):
        return self._parameters

    def get_proposal_function(self):
        params = ['mot_float_type current', 'void* rng_data']
        params.extend('mot_float_type {}'.format(p.name) for p in self._parameters)

        return '''
            #ifndef {include_guard_name}
            #define {include_guard_name}

            mot_float_type {function_name}({params}){{
                {function_body}
            }}

            #endif //{include_guard_name}
        '''.format(include_guard_name='PROPOSAL_{}'.format(self._proposal_name.upper()),
                   function_name=self.get_proposal_function_name(),
                   params=', '.join(params),
                   function_body=self._proposal_body)

    def get_proposal_function_name(self):
        return 'proposal_{}'.format(self._proposal_name)

    def get_proposal_logpdf_function(self):
        params = ['mot_float_type current', 'mot_float_type other']
        params.extend('mot_float_type {}'.format(p.name) for p in self._parameters)

        return '''
            #ifndef {include_guard_name}
            #define {include_guard_name}

            mot_float_type {function_name}({params}){{
                {function_body}
            }}

            #endif //{include_guard_name}
        '''.format(include_guard_name='PROPOSAL_LOGPDF_{}'.format(self._proposal_name.upper()),
                   function_name=self.get_proposal_logpdf_function_name(),
                   params=', '.join(params),
                   function_body=self._logpdf_body)

    def get_proposal_logpdf_function_name(self):
        return 'proposal_logpdf_{}'.format(self._proposal_name)

    def get_proposal_update_function(self):
        return self._proposal_update_function


class GaussianProposal(SimpleProposal):

    def __init__(self, std=1.0, adaptable=True, proposal_update_function=None):
        """Create a new proposal function using a Gaussian distribution with the given scale.

        Args:
            std (float): The scale of the Gaussian distribution.
            adaptable (boolean): If this proposal is adaptable during sampling
            proposal_update_function (ProposalUpdate): the proposal update function to use. Defaults
                to default in :class:`SimpleProposal`.
        """
        parameters = [ProposalParameter('std', std, adaptable)]
        super(GaussianProposal, self).__init__(
            'return fma(std, (mot_float_type)frandn(rng_data), current);',
            'gaussian',
            parameters,
            proposal_update_function=proposal_update_function
        )


class CircularGaussianProposal(SimpleProposal):

    def __init__(self, modulus, std=1.0, adaptable=True, proposal_update_function=None):
        """A Gaussian distribution which loops around the given modulus.

        Args:
            modulus (float): at which point we loop around
            std (float): The scale of the Gaussian distribution.
            adaptable (boolean): If this proposal is adaptable during sampling
            proposal_update_function (ProposalUpdate): the proposal update function to use. Defaults
                to default in :class:`SimpleProposal`.
        """
        parameters = [ProposalParameter('std', std, adaptable)]
        super(CircularGaussianProposal, self).__init__(
            '''
                double x1 = fma(std, (mot_float_type)frandn(rng_data), current);
                double x2 = {};
                return (mot_float_type) (x1 - floor(x1 / x2) * x2);
            '''.format(modulus),
            'circular_gaussian',
            parameters,
            proposal_update_function=proposal_update_function
        )



import numpy as np
from mot.cl_data_type import CLDataType
from mot.model_building.parameter_functions.proposals import GaussianProposal

__author__ = 'Robbert Harms'
__date__ = "2014-06-19"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class AbstractParameterPrior(object):
    """The priors are used during model sampling.

    These priors should be in the

    They indicate the a priori information one has about a parameter.
    """

    def get_prior_function(self):
        """Get the prior function as a CL string. This should include include guards (#ifdef's).

        This should follow the signature:

        .. code-block: c

            mot_float_type <prior_fname>(mot_float_type parent_parameter,
                                         mot_float_type lower_bound,
                                         mot_float_type upper_bound,
                                         <sub-parameters>)

        That is, the parent parameter and it lower and upper bound is given next to the optional parameters
        defined in this prior.

        Returns:
            str: The cl function
        """

    def get_prior_function_name(self):
        """Get the name of the prior function call.

         This is used by the model builder to construct the call to the prior function.

         Returns:
            str: name of the function
        """

    def get_parameters(self):
        """Get the additional parameters featured in this prior.

        This can return a list of additional parameters to be used in the model function.

        Returns:
            list of CLFunctionParameter: the list of function parameters to be added to the list of
                parameters of the enclosing model.
        """
        return []


class SimplePrior(AbstractParameterPrior):

    def __init__(self, prior_body, prior_name, prior_params=None):
        self._prior_body = prior_body
        self._prior_name = prior_name
        self._prior_params = prior_params or []

    def get_parameters(self):
        return self._prior_params

    def get_prior_function(self):
        params = ['value', 'lower_bound', 'upper_bound']
        params.extend(p.name for p in self._prior_params)
        params = ['const mot_float_type {}'.format(v) for v in params]

        return '''
            #ifndef {include_guard_name}
            #define {include_guard_name}

            mot_float_type {function_name}({params}){{
                {prior_body}
            }}

            #endif //{include_guard_name}
        '''.format(include_guard_name='PRIOR_{}'.format(self._prior_name.upper()), function_name=self._prior_name,
                   prior_body=self._prior_body, params=', '.join(params))

    def get_prior_function_name(self):
        return self._prior_name


class AlwaysOne(SimplePrior):

    def __init__(self):
        """The uniform prior is always 1. :math:`P(v) = 1` """
        super(AlwaysOne, self).__init__('return 1;', 'uniform')


class ReciprocalPrior(SimplePrior):

    def __init__(self):
        """The reciprocal of the current value. :math:`P(v) = 1/v` """
        body = '''
            if(value <= 0){
                return 0;
            }
            return 1.0/value;
        '''
        super(ReciprocalPrior, self).__init__(body, 'reciprocal')


class UniformWithinBoundsPrior(SimplePrior):

    def __init__(self):
        """This prior is 1 within the upper and lower bound of the parameter, 0 outside."""
        super(UniformWithinBoundsPrior, self).__init__(
            'return (value < lower_bound || value > upper_bound) ? 0.0 : 1.0;',
            'uniform_within_bounds')


class AbsSinPrior(SimplePrior):

    def __init__(self):
        """Angular prior: :math:`P(v) = |\\sin(v)|`"""
        super(AbsSinPrior, self).__init__('return fabs(sin(value));', 'abs_sin')


class AbsSinHalfPrior(SimplePrior):

    def __init__(self):
        """Angular prior: :math:`P(v) = |\\sin(x)/2.0|`"""
        super(AbsSinHalfPrior, self).__init__('return fabs(sin(value)/2.0);', 'abs_sin_half')


class VagueGammaPrior(SimplePrior):

    def __init__(self):
        """The vague gamma prior is meant as a proper uniform prior.

        Lee & Wagenmakers:

            The practice of assigning Gamma(0.001, 0.001) priors on precision parameters is theoretically motivated by
            scale invariance arguments, meaning that priors are chosen so that changing the measurement
            scale of the data does not affect inference.
            The invariant prior on precision λ corresponds to a uniform distribution on log σ,
            that is, p (σ2) ∝ 1/σ2, or a Gamma(a → 0, b → 0) distribution.
            This invariant prior distribution, however, is improper (i.e., the area under the curve is unbounded),
            which means it is not really a distribution, but the limit of a sequence of distributions
            (see Jaynes, 2003). WinBUGS requires the use of proper distributions,
            and the Gamma(0.001, 0.001) prior is intended as a proper approximation to the theoretically
            motivated improper prior. This raises the issue of whether inference is sensitive to the essentially
            arbitrary value 0.001, and it is sometimes the case that using other small values such as 0.01 or 0.1
            leads to more stable sampling
            in WinBUGS.

            -- Lee & Wagenmakers, Bayesian Cognitive Modeling, 2014, Chapter 4, Box 4.1

        While this is not WinBUGS and improper priors are allowed in MOT, it is still useful to have this prior
        in case people desire proper priors.
        """
        body = '''
            float kappa = 0.001;
            float theta = 1/0.001;

            return (1.0 / (tgamma(kappa) * pow(theta, kappa))) * pow(value, kappa - 1) * exp(- value / theta);
        '''
        super(VagueGammaPrior, self).__init__(body, 'vague_gamma_prior', [])


class NormalPDF(SimplePrior):

    def __init__(self):
        """Normal PDF on the given value: :math:`P(v) = N(v; \\mu, \\sigma)`"""
        from mot.model_building.cl_functions.parameters import FreeParameter
        params = [FreeParameter(CLDataType.from_string('mot_float_type'), 'mu', True, 0, -np.inf, np.inf,
                                sampling_prior=AlwaysOne()),
                  FreeParameter(CLDataType.from_string('mot_float_type'), 'sigma', False, 1, -np.inf, np.inf,
                                sampling_prior=AlwaysOne())]

        super(NormalPDF, self).__init__(
            'return exp(-pown(value - mu, 2) / (2 * pown(sigma, 2))) / (sigma * sqrt(2 * M_PI));',
            'normal_pdf',
            params)


class ARDBeta(SimplePrior):

    def __init__(self):
        """This is a collapsed form of the Beta PDF meant for use in Automatic Relevance Detection sampling.

        In this prior the ``alpha`` parameter of the Beta prior is set to 1 which simplifies the equation.
        The parameter ``beta`` is still free and can be changed as desired.

        The implemented prior is :math:`B(x; 1, \\beta) = \\beta * (1 - x)^{\\beta - 1}`

        """
        from mot.model_building.cl_functions.parameters import FreeParameter
        params = [FreeParameter(CLDataType.from_string('mot_float_type'), 'beta', False, 1, 1e-4, 1000,
                                sampling_prior=ReciprocalPrior(),
                                sampling_proposal=GaussianProposal(0.01))]

        body = '''
            if(value < 0 || value > 1){
                return 0;
            }
            return beta * pow(1 - value, beta - 1);
        '''
        super(ARDBeta, self).__init__(body, 'ard_beta_pdf', params)


class ARDGaussian(SimplePrior):

    def __init__(self):
        """This is a Gaussian prior meant for use in Automatic Relevance Detection sampling.

        This uses a Gaussian prior with mean at zero and a standard deviation determined by the ``alpha`` parameter
        with the relationship :math:`\sigma = 1/\\sqrt(\\alpha)`.
        """
        from mot.model_building.cl_functions.parameters import FreeParameter
        params = [FreeParameter(CLDataType.from_string('mot_float_type'), 'alpha', False, 8, 1e-5, 1e4,
                                sampling_prior=UniformWithinBoundsPrior(),
                                sampling_proposal=GaussianProposal(20))]

        body = '''
            if(value < 0 || value > 1){
                return 0;
            }
            mot_float_type sigma = 1.0/sqrt(alpha);
            return exp(-pown(value, 2) / (2 * pown(sigma, 2))) / (sigma * sqrt(2 * M_PI));
        '''
        super(ARDGaussian, self).__init__(body, 'ard_beta_pdf', params)



import os
from pkg_resources import resource_filename
from mot.cl_data_type import CLDataType
from mot.model_building.cl_functions.base import ModelFunction
from mot.model_building.cl_functions.parameters import FreeParameter
from mot.model_building.parameter_functions.priors import ARDGaussian, UniformWithinBoundsPrior
from mot.model_building.parameter_functions.proposals import GaussianProposal
from mot.model_building.parameter_functions.transformations import ClampTransform, CosSqrClampTransform

__author__ = 'Robbert Harms'
__date__ = "2016-10-03"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Scalar(ModelFunction):

    def __init__(self, name='Scalar', param_name='s', value=0.0, lower_bound=0.0, upper_bound=float('inf'),
                 parameter_kwargs=None):
        """A Scalar model function to be used during optimization.

        Args:
            name (str): The name of the model
            value (number or ndarray): The initial value for the single free parameter of this function.
            lower_bound (number or ndarray): The initial lower bound for the single free parameter of this function.
            upper_bound (number or ndarray): The initial upper bound for the single free parameter of this function.
            parameter_kwargs (dict): additional settings for the parameter initialization
        """
        parameter_settings = dict(parameter_transform=ClampTransform(),
                                  sampling_proposal=GaussianProposal(1.0))
        parameter_settings.update(parameter_kwargs or {})

        super(Scalar, self).__init__(
            name,
            'cmScalar',
            (FreeParameter(CLDataType.from_string('mot_float_type'), param_name,
                           False, value, lower_bound, upper_bound, **parameter_settings),))

    def get_cl_header(self):
        """See base class for details"""
        path = resource_filename('mot', 'data/opencl/model_functions/Scalar.h')
        return self._get_cl_dependency_headers() + "\n" + open(os.path.abspath(path), 'r').read()

    def get_cl_code(self):
        """See base class for details"""
        path = resource_filename('mot', 'data/opencl/model_functions/Scalar.cl')
        return self._get_cl_dependency_code() + "\n" + open(os.path.abspath(path), 'r').read()


class Weight(Scalar):

    def __init__(self, name='Weight', value=0.5, lower_bound=0.0, upper_bound=1.0, parameter_kwargs=None):
        """Implements Scalar model function to add the semantics of representing a Weight.

        Some of the code checks for type Weight, be sure to use this model function if you want to represent a Weight.

        A weight is meant to be a model volume fraction.

        Args:
            name (str): The name of the model
            value (number or ndarray): The initial value for the single free parameter of this function.
            lower_bound (number or ndarray): The initial lower bound for the single free parameter of this function.
            upper_bound (number or ndarray): The initial upper bound for the single free parameter of this function.
        """
        parameter_settings = dict(parameter_transform=CosSqrClampTransform(),
                                  sampling_proposal=GaussianProposal(0.01),
                                  sampling_prior=UniformWithinBoundsPrior())
        parameter_settings.update(parameter_kwargs or {})

        super(Weight, self).__init__(name=name, param_name='w', value=value, lower_bound=lower_bound,
                                     upper_bound=upper_bound, parameter_kwargs=parameter_settings)


class ARD_Beta_Weight(Weight):

    def __init__(self, name='ARD_Beta_Weight', value=0.5, lower_bound=0.0, upper_bound=1.0):
        """A compartment weight with a Beta prior, to be used in Automatic Relevance Detection

        It is exactly the same as a weight, except that it has a different prior, a Beta distribution prior between
        [0, 1].

        Args:
            name (str): The name of the model
            value (number or ndarray): The initial value for the single free parameter of this function.
            lower_bound (number or ndarray): The initial lower bound for the single free parameter of this function.
            upper_bound (number or ndarray): The initial upper bound for the single free parameter of this function.
        """
        parameter_settings = dict(sampling_prior=ARDGaussian())

        super(ARD_Beta_Weight, self).__init__(name=name, value=value, lower_bound=lower_bound,
                                              upper_bound=upper_bound, parameter_kwargs=parameter_settings)


class ARD_Gaussian_Weight(Weight):

    def __init__(self, name='ARD_Gaussian_Weight', value=0.5, lower_bound=0.0, upper_bound=1.0):
        """A compartment weight with a Gaussian prior, to be used in Automatic Relevance Detection

        It is exactly the same as a weight, except that it has a different prior, a Gaussian prior with mean at zero
        and std given by a hyperparameter.

        Args:
            name (str): The name of the model
            value (number or ndarray): The initial value for the single free parameter of this function.
            lower_bound (number or ndarray): The initial lower bound for the single free parameter of this function.
            upper_bound (number or ndarray): The initial upper bound for the single free parameter of this function.
        """
        parameter_settings = dict(sampling_prior=ARDGaussian())

        super(ARD_Gaussian_Weight, self).__init__(name=name, value=value, lower_bound=lower_bound,
                                                  upper_bound=upper_bound, parameter_kwargs=parameter_settings)



import math
import pyopencl as cl
import numpy as np
from mot.cl_routines.mapping.error_measures import ErrorMeasures
from mot.cl_routines.mapping.residual_calculator import ResidualCalculator
from mot.random123 import get_random123_cl_code, RandomStartingPoint
from ...utils import results_to_dict, get_float_type_def
from ...load_balance_strategies import Worker
from ...cl_routines.sampling.base import AbstractSampler


__author__ = 'Robbert Harms'
__date__ = "2014-02-05"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class MetropolisHastings(AbstractSampler):

    def __init__(self, cl_environments=None, load_balancer=None, nmr_samples=None, burn_length=None,
                 sample_intervals=None, use_adaptive_proposals=True, **kwargs):
        """An CL implementation of Metropolis Hastings.

        Args:
            cl_environments: a list with the cl environments to use
            load_balancer: the load balance strategy to use
            nmr_samples (int): The length of the (returned) chain per voxel, defaults to 500
            burn_length (int): The length of the burn in (per voxel), these are extra samples,
                jump is set to 1 (no thinning)
            sample_intervals (int): how many sample we wait before storing one.
                This will draw extra samples (chain_length * sample_intervals). If set to zero we
                store every sample after the burn in.
            use_adaptive_proposals (boolean): if we use the adaptive proposals (set to True) or not (set to False).

        Attributes:
            nmr_samples (int): The length of the (returned) chain per voxel
            burn_length (int): The length of the burn in (per voxel), these are extra samples,
                jump is set to 1 (no thinning)
            sample_intervals (int): how many samples we wait before we take,
                this will draw extra samples (chain_length * sample_intervals)
            proposal_update_intervals (int): after how many samples we would like to update the proposals
                This is during burning and sampling. A value of 1 means update after every jump.
        """
        super(MetropolisHastings, self).__init__(cl_environments=cl_environments, load_balancer=load_balancer, **kwargs)
        self._nmr_samples = nmr_samples or 500
        self.burn_length = burn_length
        self.sample_intervals = sample_intervals
        self.use_adaptive_proposals = use_adaptive_proposals

        if self.burn_length is None:
            self.burn_length = 500

        if self.sample_intervals is None:
            self.sample_intervals = 0

        if self.burn_length < 0:
            raise ValueError('The burn length can not be smaller than 0, {} given'.format(self.burn_length))
        if self.sample_intervals < 0:
            raise ValueError('The sampling interval can not be smaller than 0, {} given'.format(self.sample_intervals))
        if self._nmr_samples < 1:
            raise ValueError('The number of samples to draw can '
                             'not be smaller than 1, {} given'.format(self._nmr_samples))


    @property
    def nmr_samples(self):
        return self._nmr_samples

    @nmr_samples.setter
    def nmr_samples(self, value):
        self._nmr_samples = value or 1

    def sample(self, model, init_params=None, full_output=False):
        np_dtype = np.float32
        if model.double_precision:
            np_dtype = np.float64

        self._do_initial_logging(model)

        parameters = model.get_initial_parameters(init_params)

        samples = np.zeros((model.get_nmr_problems(), parameters.shape[1], self.nmr_samples),
                           dtype=np_dtype, order='C')

        proposal_state = np.zeros((model.get_nmr_problems(), len(model.get_proposal_state())),
                                  dtype=np_dtype, order='C')

        self._logger.info('Starting sampling with method {0}'.format(self.__class__.__name__))

        workers = self._create_workers(lambda cl_environment: _MHWorker(
            cl_environment, self.get_compile_flags_list(), model, parameters, samples, proposal_state,
            self.nmr_samples, self.burn_length, self.sample_intervals, self.use_adaptive_proposals))
        self.load_balancer.process(workers, model.get_nmr_problems())

        samples_dict = results_to_dict(samples, model.get_optimized_param_names())

        self._logger.info('Finished sampling')

        if full_output:
            self._logger.info('Starting post-sampling transformations')
            volume_maps = model.finalize_optimization_results(model.samples_to_statistics(samples_dict))
            self._logger.info('Finished post-sampling transformations')

            self._logger.info('Calculating errors measures')
            errors = ResidualCalculator(cl_environments=self.cl_environments,
                                        load_balancer=self.load_balancer).calculate(model, volume_maps)
            error_measures = ErrorMeasures(self.cl_environments, self.load_balancer,
                                           model.double_precision).calculate(errors)
            volume_maps.update(error_measures)
            self._logger.info('Done calculating errors measures')

            proposal_state_dict = results_to_dict(proposal_state, model.get_proposal_state_names())

            return samples_dict, volume_maps, proposal_state_dict

        return samples_dict

    def _do_initial_logging(self, model):
        self._logger.info('Entered sampling routine.')
        self._logger.info('We will use a {} precision float type for the calculations.'.format(
            'double' if model.double_precision else 'single'))

        for env in self.load_balancer.get_used_cl_environments(self.cl_environments):
            self._logger.info('Using device \'{}\'.'.format(str(env)))

        self._logger.debug('Using compile flags: {}'.format(self.get_compile_flags_list()))

        self._logger.info('The parameters we will sample are: {0}'.format(model.get_optimized_param_names()))

        sample_settings = dict(nmr_samples=self.nmr_samples,
                               burn_length=self.burn_length,
                               sample_intervals=self.sample_intervals,
                               use_adaptive_proposals=self.use_adaptive_proposals)
        self._logger.info('Sample settings: nmr_samples: {nmr_samples}, burn_length: {burn_length}, '
                          'sample_intervals: {sample_intervals}, '
                          'use_adaptive_proposals: {use_adaptive_proposals}. '.format(**sample_settings))

        samples_drawn = dict(samples_drawn=(self.burn_length + (self.sample_intervals + 1) * self.nmr_samples),
                             samples_returned=self.nmr_samples)
        self._logger.info('Total samples drawn: {samples_drawn}, total samples returned: '
                          '{samples_returned} (per problem).'.format(**samples_drawn))


class _MHWorker(Worker):

    def __init__(self, cl_environment, compile_flags, model, parameters, samples, proposal_state,
                 nmr_samples, burn_length, sample_intervals, use_adaptive_proposals):
        super(_MHWorker, self).__init__(cl_environment)

        self._model = model
        self._parameters = parameters
        self._nmr_params = parameters.shape[1]
        self._samples = samples
        self._proposal_state = proposal_state

        self._nmr_samples = nmr_samples
        self._burn_length = burn_length
        self._sample_intervals = sample_intervals
        self.use_adaptive_proposals = use_adaptive_proposals

        self._workgroup_size = min(cl_environment.device.max_work_group_size // 4,
                                   max(1, 2**math.floor(math.log(self._model.get_nmr_inst_per_problem(), 2))))

        self._rand123_starting_point = RandomStartingPoint()

        self._kernel = self._build_kernel(compile_flags)

    def calculate(self, range_start, range_end):
        nmr_problems = range_end - range_start

        data_buffers = [cl.Buffer(self._cl_run_context.context,
                                  cl.mem_flags.READ_ONLY | cl.mem_flags.USE_HOST_PTR,
                                  hostbuf=self._parameters[range_start:range_end, :])]
        samples_buf = cl.Buffer(self._cl_run_context.context,
                                cl.mem_flags.WRITE_ONLY | cl.mem_flags.USE_HOST_PTR,
                                hostbuf=self._samples[range_start:range_end, ...])
        data_buffers.append(samples_buf)

        proposal_buffer = cl.Buffer(self._cl_run_context.context,
                                    cl.mem_flags.WRITE_ONLY | cl.mem_flags.USE_HOST_PTR,
                                    hostbuf=self._proposal_state[range_start:range_end, ...])
        data_buffers.append(proposal_buffer)

        for data in self._model.get_data():
            data_buffers.append(cl.Buffer(self._cl_run_context.context,
                                          cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR, hostbuf=data))

        kernel_event = self._kernel.sample(self._cl_run_context.queue,
                                           (int(nmr_problems * self._workgroup_size),),
                                           (int(self._workgroup_size),),
                                           *data_buffers)

        return [self._enqueue_readout(samples_buf, self._samples, 0, nmr_problems, [kernel_event]),
                self._enqueue_readout(proposal_buffer, self._proposal_state, 0, nmr_problems, [kernel_event])]

    def _get_kernel_source(self):
        kernel_param_names = ['global mot_float_type* params',
                              'global mot_float_type* samples',
                              'global mot_float_type* final_proposal_state']
        kernel_param_names.extend(self._model.get_kernel_param_names(self._cl_environment.device))

        proposal_state_size = len(self._model.get_proposal_state())

        kernel_source = '''
            #define NMR_INST_PER_PROBLEM ''' + str(self._model.get_nmr_inst_per_problem()) + '''
        '''
        kernel_source += get_random123_cl_code()
        kernel_source += get_float_type_def(self._model.double_precision)
        kernel_source += self._model.get_kernel_data_struct(self._cl_environment.device)
        kernel_source += self._model.get_log_prior_function('getLogPrior', address_space_parameter_vector='local')
        kernel_source += self._model.get_proposal_function('getProposal', address_space_proposal_state='local')
        kernel_source += self._model.get_log_likelihood_per_observation_function('getLogLikelihoodPerObservation',
                                                                                 full_likelihood=False)

        if self.use_adaptive_proposals:
            kernel_source += self._model.get_proposal_state_update_function('updateProposalState',
                                                                            address_space='local')

        if not self._model.is_proposal_symmetric():
            kernel_source += self._model.get_proposal_logpdf('getProposalLogPDF', address_space_proposal_state='local')

        kernel_source += '''
            void _fill_log_likelihood_tmp(const void* const data,
                                          local mot_float_type* const x_local,
                                          local double* log_likelihood_tmp){

                int observation_ind;
                uint local_id = get_local_id(0);
                log_likelihood_tmp[local_id] = 0;

                mot_float_type x_private[''' + str(self._nmr_params) + '''];
                for(int i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                    x_private[i] = x_local[i];
                }

                for(int i = 0; i < ceil(NMR_INST_PER_PROBLEM /
                                        (mot_float_type)''' + str(self._workgroup_size) + '''); i++){

                    observation_ind = i * ''' + str(self._workgroup_size) + ''' + local_id;

                    if(observation_ind < NMR_INST_PER_PROBLEM){
                        log_likelihood_tmp[local_id] += getLogLikelihoodPerObservation(
                            data, x_private, observation_ind);
                    }
                }

                barrier(CLK_LOCAL_MEM_FENCE);
            }

            void _sum_log_likelihood_tmp_local(local double* log_likelihood_tmp, local double* log_likelihood){
                *log_likelihood = 0;
                for(int i = 0; i < ''' + str(self._workgroup_size) + '''; i++){
                    *log_likelihood += log_likelihood_tmp[i];
                }
            }

            void _sum_log_likelihood_tmp_private(local double* log_likelihood_tmp, private double* log_likelihood){
                *log_likelihood = 0;
                for(int i = 0; i < ''' + str(self._workgroup_size) + '''; i++){
                    *log_likelihood += log_likelihood_tmp[i];
                }
            }
        '''

        kernel_source += '''
            void _update_state(local mot_float_type* const x_local,
                               void* rng_data,
                               local double* const current_likelihood,
                               local mot_float_type* const current_prior,
                               const void* const data,
                               local mot_float_type* const proposal_state,
                               local uint * const acceptance_counter,
                               local double* log_likelihood_tmp){

                float4 random_nmr;
                mot_float_type new_prior = 0;
                double new_likelihood;
                double bayesian_f;
                mot_float_type old_x;
                bool is_first_work_item = get_local_id(0) == 0;

                #pragma unroll 1
                for(int k = 0; k < ''' + str(self._nmr_params) + '''; k++){
                    if(is_first_work_item){
                        random_nmr = frand(rng_data);

                        old_x = x_local[k];
                        x_local[k] = getProposal(k, x_local[k], rng_data, proposal_state);

                        new_prior = getLogPrior(data, x_local);
                    }

                    if(exp(new_prior) > 0){
                        _fill_log_likelihood_tmp(data, x_local, log_likelihood_tmp);

                        if(is_first_work_item){
                            _sum_log_likelihood_tmp_private(log_likelihood_tmp, &new_likelihood);

        '''
        if self._model.is_proposal_symmetric():
            kernel_source += '''
                            bayesian_f = exp((new_likelihood + new_prior) - (*current_likelihood + *current_prior));
                '''
        else:
            kernel_source += '''
                            mot_float_type x_to_prop = getProposalLogPDF(k, old_x, x_local[k], proposal_state);
                            mot_float_type prop_to_x = getProposalLogPDF(k, x_local[k], x_local[k], proposal_state);

                            bayesian_f = exp((new_likelihood + new_prior + x_to_prop) -
                                                (*current_likelihood + *current_prior + prop_to_x));
                '''
        kernel_source += '''

                            if(random_nmr.x < bayesian_f){
                                *current_likelihood = new_likelihood;
                                *current_prior = new_prior;
                                acceptance_counter[k]++;
                            }
                            else{
                                x_local[k] = old_x;
                            }
                        }
                    }
                    else{ // prior returned 0
                        if(is_first_work_item){
                            x_local[k] = old_x;
                        }
                    }
                }
            }
        '''
        kernel_source += '''
            void _sample(local mot_float_type* const x_local,
                         void* rng_data,
                         local double* const current_likelihood,
                         local mot_float_type* const current_prior,
                         const void* const data,
                         local mot_float_type* const proposal_state,
                         local uint* const sampling_counter,
                         local uint* const acceptance_counter,
                         local mot_float_type* parameter_mean,
                         local mot_float_type* parameter_variance,
                         local mot_float_type* parameter_m2,
                         global mot_float_type* samples,
                         local double* log_likelihood_tmp){

                uint i, j;
                uint problem_ind = get_group_id(0);
                bool is_first_work_item = get_local_id(0) == 0;
                mot_float_type previous_mean;

                for(j = 0; j < ''' + str(self._nmr_params) + '''; j++){
                    parameter_mean[j] = x_local[j];
                    parameter_variance[j] = 0;
                    parameter_m2[j] = 0;
                }

                for(i = 0; i < ''' + str(self._nmr_samples * (self._sample_intervals + 1)
                                         + self._burn_length) + '''; i++){

                    _update_state(x_local, rng_data, current_likelihood, current_prior,
                                  data, proposal_state, acceptance_counter,
                                  log_likelihood_tmp);

                    if(is_first_work_item){
                        for(j = 0; j < ''' + str(self._nmr_params) + '''; j++){
                            sampling_counter[j]++;
                        }

                        ''' + ('updateProposalState(proposal_state, sampling_counter, acceptance_counter, '
                               'parameter_variance);'
                               if self.use_adaptive_proposals else '') + '''

                        if(i > 0){
                            for(j = 0; j < ''' + str(self._nmr_params) + '''; j++){
                                // online variance, algorithm by Welford
                                //  B. P. Welford (1962)."Note on a method for calculating corrected sums of squares
                                //      and products". Technometrics 4(3):419–420.
                                //
                                // and studied in:
                                // Chan, Tony F.; Golub, Gene H.; LeVeque, Randall J. (1983).
                                //      Algorithms for Computing the Sample Variance: Analysis and Recommendations.
                                //      The American Statistician 37, 242-247. http://www.jstor.org/stable/2683386

                                previous_mean = parameter_mean[j];
                                parameter_mean[j] += (x_local[j] - parameter_mean[j]) / (i + 1);
                                parameter_m2[j] += (x_local[j] - previous_mean) * (x_local[j] - parameter_mean[j]);

                                if(i > 1){
                                    parameter_variance[j] = parameter_m2[j] / (i - 1);
                                }
                            }
                        }

                        if(i >= ''' + str(self._burn_length) + '''
                           && i % ''' + str(self._sample_intervals + 1) + ''' == 0){

                            for(j = 0; j < ''' + str(self._nmr_params) + '''; j++){
                                samples[(uint)((i - ''' + str(self._burn_length) + ''') // remove the burn-in
                                                / ''' + str(self._sample_intervals + 1) + ''') // remove the interval
                                        + j * ''' + str(self._nmr_samples) + '''  // parameter index
                                        + problem_ind * ''' + str(self._nmr_params * self._nmr_samples) + ''']
                                            = x_local[j];
                            }
                        }
                    }
                }
            }
        '''

        kernel_source += '''
            __kernel void sample(
                ''' + ",\n".join(kernel_param_names) + '''
                ){

                uint problem_ind = get_group_id(0);

                ''' + self._model.get_kernel_data_struct_initialization(self._cl_environment.device,
                                                                        'data', 'problem_ind') + '''

                rand123_data rand123_rng_data = ''' + self._get_rand123_init_cl_code() + ''';
                void* rng_data = (void*)&rand123_rng_data;

                local mot_float_type x_local[''' + str(self._nmr_params) + '''];

                local double current_likelihood;
                local mot_float_type current_prior;

                local mot_float_type proposal_state[''' + str(proposal_state_size) + '''];

                // the following items are used for updating the proposals, do not use them for other purposes
                local uint sampling_counter[''' + str(self._nmr_params) + '''];
                local uint acceptance_counter[''' + str(self._nmr_params) + '''];
                local mot_float_type parameter_mean[''' + str(self._nmr_params) + '''];
                local mot_float_type parameter_m2[''' + str(self._nmr_params) + ''']; // used in calculation of variance
                local mot_float_type parameter_variance[''' + str(self._nmr_params) + '''];

                local double log_likelihood_tmp[''' + str(self._workgroup_size) + '''];

                if(get_local_id(0) == 0){
                    for(int i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                        x_local[i] = params[problem_ind * ''' + str(self._nmr_params) + ''' + i];
                    }

                    current_prior = getLogPrior((void*)&data, x_local);

                    ''' + ' '.join('proposal_state[{}] = {};'.format(i, v)
                                   for i, v in enumerate(self._model.get_proposal_state())) + '''
                    ''' + ' '.join('sampling_counter[{}] = 0;'.format(i)
                                   for i in range(self._nmr_params)) + '''
                    ''' + ' '.join('acceptance_counter[{}] = 0;'.format(i)
                                   for i in range(self._nmr_params)) + '''
                }
                _fill_log_likelihood_tmp((void*)&data, x_local, log_likelihood_tmp);
                _sum_log_likelihood_tmp_local(log_likelihood_tmp, &current_likelihood);

                _sample(x_local, rng_data, &current_likelihood,
                    &current_prior, (void*)&data, proposal_state, sampling_counter, acceptance_counter,
                    parameter_mean, parameter_variance, parameter_m2, samples, log_likelihood_tmp);

                if(get_local_id(0) == 0){
                    for(int i = 0; i < ''' + str(proposal_state_size) + '''; i++){
                        final_proposal_state[problem_ind
                                             * ''' + str(proposal_state_size) + ''' + i] = proposal_state[i];
                    }
                }
            }
        '''
        return kernel_source

    def _get_rand123_init_cl_code(self):
        key = self._rand123_starting_point.get_key()
        counter = self._rand123_starting_point.get_counter()

        if len(key):
            return 'rand123_initialize_data_extra_precision((uint[]){%(c0)r, %(c1)r, %(c2)r, %(c3)r}, ' \
                   '(uint[]){%(k0)r, %(k1)r})' % {'c0': counter[0], 'c1': counter[1],
                                                  'c2': counter[2], 'c3': counter[3],
                                                  'k0': key[0], 'k1': counter[1]}
        else:
            return 'rand123_initialize_data((uint[]){%(c0)r, %(c1)r, %(c2)r, %(c3)r})' \
                   % {'c0': counter[0], 'c1': counter[1], 'c2': counter[2], 'c3': counter[3]}

